// DelaunayTriangulation.cpp


//#define DEBUG_STEP

static bool g_doDebugStep = true;

#include <math.h>

#include "DelaunayTriangulation.h"


static float EPSILONF = 0.001f;


#ifdef DEBUG_STEP
int g_debugStep = 0;
#endif

struct stTris
{
	stTris() {i0=i1=i2=-1;  valid=true; }
	int i0, i1, i2;
	bool valid;
};

struct stEdge
{
	stEdge(){i0=i1=-1;}
	int i0, i1;
};


/******************************************************************************/


float Cross(const D3DXVECTOR2& p0, const D3DXVECTOR2& p1)
{
	return p0.y * p1.x - p0.x * p1.y;
}

/******************************************************************************/


bool SameSign(const float a, const float b)
{
	if ( (a * b) > 0.0f ) return true;
	return false;
}


/******************************************************************************/

bool IsPointInTriangle(const D3DXVECTOR2& pA,
					   const D3DXVECTOR2& pB,
					   const D3DXVECTOR2& pC,
					   const D3DXVECTOR2& point)
{
	float pab = Cross(point-pA, pB-pA);
	float pbc = Cross(point-pB, pC-pB);

	if (!SameSign(pab, pbc)) return false;

	float pca = Cross(point-pC, pA-pC);

	if (!SameSign(pab, pca)) return false;

	return true;
}

/******************************************************************************/

bool LineLineIntersect( const D3DXVECTOR2& from0,
					    const D3DXVECTOR2& to0,
					    const D3DXVECTOR2& from1,
					    const D3DXVECTOR2& to1,
					    D3DXVECTOR2* hitPoint=NULL)
{
	float x1 = from0.x;
	float y1 = from0.y;
	float x2 = to0.x;
	float y2 = to0.y;
	float x3 = from1.x;
	float y3 = from1.y;
	float x4 = to1.x;
	float y4 = to1.y;

	float denom = (y4 - y3)*(x2 - x1) - (x4 - x3)*(y2 - y1);

	// If denom is zero the lines are parallel
	if ( fabs(denom)<0.00001f ) return false;

	float t0 = ((x4 - x3)*(y1 - y3) - (y4 - y3)*(x1 - x3)) / denom;
		       
	float t1 = ((x2 - x1)*(y1 - y3) - (y2 - y1)*(x1 - x3)) / denom;

	//if ( fabs(t0)<0.00000001f ) return false;
	//if ( fabs(t1)<0.00000001f ) return false;

	//static float thresh0 = -0.000000001f;
	float thresh0 = 0.0001f;
	static float thresh1 = 0.0f;
	if ( t0<=thresh0 || t0>=(1.0f-thresh0) || t1 <=thresh1 || t1>=(1.0f-thresh1) )
	{
		return false;
	}

	if (hitPoint)
	{
		(*hitPoint) = from0 + (to0-from0)*t0;
	}

	return true;
}

/******************************************************************************/

bool CircumCircle(float xp, float yp, 
				  float x1, float y1, 
				  float x2, float y2, 
				  float x3, float y3, 
				  float &xc, float &yc, float &r)
{
	 float m1, m2, mx1, mx2, my1, my2;
	 float dx, dy, rsqr, drsqr;

	 static float COLINEAR_MIN = 0.000088f;
	 /* Check for coincident points */
	 
	 if(fabs(y1 - y2) < COLINEAR_MIN && fabs(y2 - y3) < COLINEAR_MIN)
	 {
		 return(false);
	 }
	 if(fabs(x1 - x2) < COLINEAR_MIN && fabs(x2 - x3) < COLINEAR_MIN)
	 {
		 return(false);
	 }
	 

	 //if (  fabs( x1*(y2-y3) + x2*(y3-y1) + x3*(y1-y2) ) < COLINEAR_MIN )
	 //{
		// return false;
	 //}

	 if(fabs(y2-y1) < EPSILONF)
	 { 
		 m2 = - (x3 - x2) / (y3 - y2);
		 mx2 = (x2 + x3) * 0.5f;
		 my2 = (y2 + y3) * 0.5f;
		 xc = (x2 + x1)  * 0.5f;
		 yc = m2 * (xc - mx2) + my2;
	 }
	 else if(fabs(y3 - y2) < EPSILONF)
	 { 
		 m1 = - (x2 - x1) / (y2 - y1);
		 mx1 = (x1 + x2) * 0.5f;
		 my1 = (y1 + y2) * 0.5f;
		 xc = (x3 + x2) * 0.5f;
		 yc = m1 * (xc - mx1) + my1;
	 }
	 else
	 {
		 m1 = - (x2 - x1) / (y2 - y1); 
		 m2 = - (x3 - x2) / (y3 - y2); 
		 mx1 = (x1 + x2) * 0.5f; 
		 mx2 = (x2 + x3) * 0.5f;
		 my1 = (y1 + y2) * 0.5f;
		 my2 = (y2 + y3) * 0.5f;
		 xc = (m1 * mx1 - m2 * mx2 + my2 - my1) / (m1 - m2); 
		 yc = m1 * (xc - mx1) + my1; 
	 }
	 dx = x2 - xc;
	 dy = y2 - yc;
	 rsqr = dx * dx + dy * dy;
	 r = sqrtf(rsqr); 
	 dx = xp - xc;
	 dy = yp - yc;
	 drsqr = dx * dx + dy * dy;
	 //static float tweek = -0.1f;
	 //return((drsqr <= rsqr-tweek) ? true : false);

	 static float tweek = 1.000001f;
	 return((drsqr <= (rsqr*tweek)) ? true : false);
}


/******************************************************************************/

int TriangulatePoints( const D3DXVECTOR2* inVerts,			// in
					   const int numVerts,					// in
					   const int maxOutTris,				// in
					   int*  outNumTris,					// out
					   int*  outTriIndices)					// out
{
	// Allocate enough memory for all our triangles
	int maxTris = numVerts * 2000;
	stTris* tris = new stTris[maxTris + 100];
	int numTris = 0;

	int maxEdges = maxTris * 3;
	stEdge* edges = new stEdge[maxEdges + 100];
	int numEdges = 0;

	//
	// Create a bit temp array to hold are verts plus 3 extra ones for our super
	// triangle
	//
	D3DXVECTOR2* verts = new D3DXVECTOR2[numVerts + 3];
	memcpy(verts, inVerts, numVerts * sizeof(D3DXVECTOR2));

	//
	// Find the maximum and minimum vertex bounds.
	// This is to allow calculation of the bounding triangle
	//
	D3DXVECTOR2 vmin = verts[0];
	D3DXVECTOR2 vmax = vmin;
	for(int i = 1; i<numVerts; i++)
	{
		vmin.x = min(vmin.x, verts[i].x);
		vmin.y = min(vmin.y, verts[i].y);
		vmax.x = max(vmax.x, verts[i].x);
		vmax.y = max(vmax.y, verts[i].y);
	}
	D3DXVECTOR2 delta = vmax - vmin;
	float dmax = (delta.x > delta.y) ? delta.x : delta.y;
	D3DXVECTOR2 mid = vmin + (vmax - vmin) * 0.5f;

	//
	// Set up the supertriangle
	// his is a triangle which encompasses all the sample points.
	// The supertriangle coordinates are added to the end of the
	// vertex list. The supertriangle is the first triangle in
	// the triangle list.
	//
	verts[numVerts+0].x = mid.x - 10 * dmax;
	verts[numVerts+0].y = mid.y - dmax;
	verts[numVerts+1].x = mid.x;
	verts[numVerts+1].y = mid.y + 10 * dmax;
	verts[numVerts+2].x = mid.x + 10 * dmax;
	verts[numVerts+2].y = mid.y - dmax;
	tris[0].i0 = numVerts + 0;
	tris[0].i1 = numVerts + 1;
	tris[0].i2 = numVerts + 2;
	numTris = 1;

	//
	// Include each point one at a time into the existing mesh
	//
	for(int i=0; i<numVerts; i++)
	{
		const D3DXVECTOR2& vert = verts[i];
		numEdges = 0;


		const D3DXVECTOR2& vertNext = verts[(i+1)%numVerts];
		const float deltaLen = D3DXVec2LengthSq( &(vertNext - vert) );
		static float minDelta = 0.00000001f;
		if (deltaLen < minDelta ) continue;


		//
		// Set up the edge buffer.
		// If the point (xp,yp) lies inside the circumcircle then the
		// three edges of that triangle are added to the edge buffer
		// and that triangle is removed.
		//
		for(int k=0; k<numTris; k++)
		{
			D3DXVECTOR2 p0 = verts[ tris[k].i0 ];
			D3DXVECTOR2 p1 = verts[ tris[k].i1 ];
			D3DXVECTOR2 p2 = verts[ tris[k].i2 ];

			float xc, yc, r;
			bool inside = CircumCircle(vert.x, vert.y,	// Test Vert
									   p0.x,   p0.y,	// 3 Corners of the triangle
									   p1.x,   p1.y, 
									   p2.x,   p2.y, 
									   xc, yc, r);		// Results

			if (inside)
			{
				DBG_ASSERT(numEdges+3 < maxEdges);
				edges[numEdges+0].i0 = tris[k].i0;
				edges[numEdges+0].i1 = tris[k].i1;
				edges[numEdges+1].i0 = tris[k].i1;
				edges[numEdges+1].i1 = tris[k].i2;
				edges[numEdges+2].i0 = tris[k].i2;
				edges[numEdges+2].i1 = tris[k].i0;
				numEdges += 3;
				DBG_ASSERT(numTris>0);

				tris[k] = tris[numTris-1]; // Remove this triangle from the list
				numTris--;
				k--;
			}
		}

		for(int j=0; j<numEdges-1; j++)
		{
			for(int k = j+1; k<numEdges; k++)
			{
				if((edges[j].i0 == edges[k].i1) && (edges[j].i1 == edges[k].i0))
				{
					edges[j].i0 = -1;
					edges[j].i1 = -1;
					edges[k].i0 = -1;
					edges[k].i1 = -1;
				}

				if ((edges[j].i0 == edges[k].i0) && (edges[j].i1 == edges[k].i1)) 
				{
					edges[j].i0 = -1;
					edges[j].i1 = -1;
					edges[k].i0 = -1;
					edges[k].i1 = -1;
				}
              
			}
		}

		//
		// Form new triangles for the current point
		// Skipping over any tagged edges.
		// All edges are arranged in clockwise order.
		//
		for(int k=0; k<numEdges; k++) 
		{
			if (edges[k].i0<0 || edges[k].i1<0) continue;

			DBG_ASSERT(numTris+1<maxTris);
			tris[numTris].i0 = edges[k].i0;
			tris[numTris].i1 = edges[k].i1;
			tris[numTris].i2 = i;
			numTris++;
		}

	}

	//
	// Remove triangles with supertriangle vertices
	// These are triangles which have a vertex number greater than nv
	//
	for(int i=0; i<numTris; i++) 
	{
		if(tris[i].i0 >= numVerts || tris[i].i1 >= numVerts || tris[i].i2 >= numVerts) 
		{
			tris[i] = tris[numTris-1]; // Remove this triangle from the list
			numTris--;
			i--;
		}
	}

	int count = 0;
	(*outNumTris) = 0;
	for (int i=0; i<numTris; i++)
	{
		DBG_ASSERT(count+3 < maxOutTris);
		if (count+3 >= maxOutTris) break;

		int i0 = tris[i].i0;
		int i1 = tris[i].i1;
		int i2 = tris[i].i2;

		outTriIndices[count]=i0;		count++;
		outTriIndices[count]=i1;		count++;
		outTriIndices[count]=i2;		count++;
		(*outNumTris) += 1;
	}

	delete[] verts;
	delete[] tris;
	delete[] edges;

	return 0;
}

/******************************************************************************/

//int XYZCompare(const void *v1, const void *v2)
//{
//		D3DXVECTOR2 *p1, *p2;
//
//		p1 = (D3DXVECTOR2*)v1;
//		p2 = (D3DXVECTOR2*)v2;
//		if(p1->x < p2->x)
//			return(-1);
//		else if(p1->x > p2->x)
//			return(1);
//		else
//			return(0);
//}

/******************************************************************************/


int TriangulatePolygon(const int numContours,				// in
					   const int*numPntsInEachContour,		// in
					   const D3DXVECTOR2* inVerts,			// in
					   const int maxOutTris,				// in
					   int*  outNumTris,					// out
					   int*  outTriIndices)					// out
{
	DBG_ASSERT(numContours>0);
	DBG_ASSERT(numPntsInEachContour);
	
	int totalVerts = 0;
	for (int i=0; i<numContours; i++) totalVerts += numPntsInEachContour[i];
	DBG_ASSERT(totalVerts>0);


	//qsort((void*)inVerts, totalVerts, sizeof(D3DXVECTOR2), XYZCompare);



#ifdef DEBUG_STEP
	extern bool g_keys[0xffff];
	if (g_keys[VK_RIGHT]) { g_keys[VK_RIGHT]=false; g_debugStep++; }
	if (g_keys[VK_LEFT])  { g_keys[VK_LEFT] =false; g_debugStep--; }
	g_debugStep = max(0, g_debugStep);
	g_debugStep = min(totalVerts, g_debugStep);

	if (g_keys[VK_HOME]) { g_keys[VK_HOME]=false; g_doDebugStep=!g_doDebugStep; }
#endif 

	// Allocate enough memory for all our triangles
	int maxTris = totalVerts * 55;
	stTris* tris = new stTris[maxTris + 100];
	int numTris = 0;

	int maxEdges = maxTris * 3;
	stEdge* edges = new stEdge[maxEdges + 100];
	int numEdges = 0;

	//
	// Create a bit temp array to hold are verts plus 3 extra ones for our super
	// triangle
	//
	D3DXVECTOR2* verts = new D3DXVECTOR2[totalVerts + 3];
	memcpy(verts, inVerts, totalVerts * sizeof(D3DXVECTOR2));


	//
	// Find the maximum and minimum vertex bounds.
	// This is to allow calculation of the bounding triangle
	//
	D3DXVECTOR2 vmin = verts[0];
	D3DXVECTOR2 vmax = vmin;
	for(int i = 1; i<totalVerts; i++)
	{
		vmin.x = min(vmin.x, verts[i].x);
		vmin.y = min(vmin.y, verts[i].y);
		vmax.x = max(vmax.x, verts[i].x);
		vmax.y = max(vmax.y, verts[i].y);
	}
	D3DXVECTOR2 delta = vmax - vmin;
	float dmax = (delta.x > delta.y) ? delta.x : delta.y;
	D3DXVECTOR2 mid = vmin + (vmax - vmin) * 0.5f;

	//
	// Set up the supertriangle
	// his is a triangle which encompasses all the sample points.
	// The supertriangle coordinates are added to the end of the
	// vertex list. The supertriangle is the first triangle in
	// the triangle list.
	//
	
	verts[totalVerts+0].x = mid.x - 4 * dmax;
	verts[totalVerts+0].y = mid.y - dmax;
	verts[totalVerts+1].x = mid.x;
	verts[totalVerts+1].y = mid.y + 4 * dmax;
	verts[totalVerts+2].x = mid.x + 4 * dmax;
	verts[totalVerts+2].y = mid.y - 4 * dmax;
	tris[0].i0 = totalVerts + 0;
	tris[0].i1 = totalVerts + 1;
	tris[0].i2 = totalVerts + 2;
	numTris = 1;


	/*
	verts[totalVerts+0].x = -5000.0f;
	verts[totalVerts+0].y = -5000.0f;
	verts[totalVerts+1].x =  0;
	verts[totalVerts+1].y =  5000.0f;
	verts[totalVerts+2].x =  5000.0f;
	verts[totalVerts+2].y = -5000.0f;
	*/
	


	//
	// Include each point one at a time into the existing mesh
	//
	for(int i=0; i<totalVerts; i++)
	{
		const D3DXVECTOR2& vert = verts[i];


		//if (i>0)
		//{
		//	static float minDelta = 0.00000001f;
		//	float delta = D3DXVec2LengthSq( &(verts[i-1]-verts[i]) );
		//	if (delta <minDelta)
		//	{
		//		continue;
		//	}
		//}

		const D3DXVECTOR2& vertNext = verts[(i+1)%totalVerts];
		const float deltaLen = D3DXVec2LengthSq( &(vertNext - vert) );
		static float minDelta = 0.00000001f;
		if (deltaLen < minDelta ) continue;

		numEdges = 0;

		//
		// Set up the edge buffer.
		// If the point (xp,yp) lies inside the circumcircle then the
		// three edges of that triangle are added to the edge buffer
		// and that triangle is removed.
		//
		for(int k=0; k<numTris; k++)
		{
			if (!tris[k].valid) continue;

			D3DXVECTOR2 p0 = verts[ tris[k].i0 ];
			D3DXVECTOR2 p1 = verts[ tris[k].i1 ];
			D3DXVECTOR2 p2 = verts[ tris[k].i2 ];

			
			const float deltaLen0 = D3DXVec2LengthSq( &(p0 - vert) );
			const float deltaLen1 = D3DXVec2LengthSq( &(p1 - vert) );
			const float deltaLen2 = D3DXVec2LengthSq( &(p2 - vert) );
			if (deltaLen0 < minDelta ) continue;
			if (deltaLen1 < minDelta ) continue;
			if (deltaLen2 < minDelta ) continue;
			


			float xc, yc, r;
			bool inside = CircumCircle(vert.x, vert.y,	// Test Vert
									   p0.x,   p0.y,	// 3 Corners of the triangle
									   p1.x,   p1.y, 
									   p2.x,   p2.y, 
									   xc, yc, r);		// Results

			if (inside)
			{
				DBG_ASSERT(numEdges+3 < maxEdges);
				edges[numEdges+0].i0 = tris[k].i0;
				edges[numEdges+0].i1 = tris[k].i1;
				edges[numEdges+1].i0 = tris[k].i1;
				edges[numEdges+1].i1 = tris[k].i2;
				edges[numEdges+2].i0 = tris[k].i2;
				edges[numEdges+2].i1 = tris[k].i0;
				numEdges += 3;
				DBG_ASSERT(numTris>0);

				tris[k] = tris[numTris-1]; // Remove this triangle from the list
				numTris--;
				k--;
				//tris[k].valid = false;
			}
		}

		for(int j=0; j<numEdges-1; j++)
		{
			for(int k = j+1; k<numEdges; k++)
			{
				if((edges[j].i0 == edges[k].i1) && (edges[j].i1 == edges[k].i0))
				{
					edges[j].i0 = -1;
					edges[j].i1 = -1;
					edges[k].i0 = -1;
					edges[k].i1 = -1;
				}

				if ((edges[j].i0 == edges[k].i0) && (edges[j].i1 == edges[k].i1)) 
				{
					edges[j].i0 = -1;
					edges[j].i1 = -1;
					edges[k].i0 = -1;
					edges[k].i1 = -1;
				}
			}
		}



		//
		// Form new triangles for the current point
		// Skipping over any tagged edges.
		// All edges are arranged in clockwise order.
		//
		if (1)
		for(int k=0; k<numEdges; k++) 
		{
			if (edges[k].i0<0 || edges[k].i1<0) continue;
		
			DBG_ASSERT(numTris+1<maxTris);
			tris[numTris].i0 = edges[k].i0;
			tris[numTris].i1 = edges[k].i1;
			tris[numTris].i2 = i;
			numTris++;
		}


#ifdef DEBUG_STEP
		if (g_doDebugStep)
		{
			extern float		g_zoom;
			extern D3DXVECTOR2	g_off;
			float s = g_zoom;
			extern LPDIRECT3DDEVICE9	g_pD3DDevice;
			D3DXVECTOR2 pnt = vert*s + g_off;
			unsigned int col = 0xEFE1FF44;
			if (g_debugStep==i)
			{
				col = 0xff00ff00;
				DebugWireDrawBox2D(g_pD3DDevice, pnt.x, pnt.y, 5, 5, col);
			}
			DebugWireDrawBox2D(g_pD3DDevice, pnt.x, pnt.y, 5, 5, col);
		}
#endif

	}

	static bool removeSuperTris = true;
#ifdef DEBUG_STEP
	if (g_keys[VK_SPACE]) { g_keys[VK_SPACE]=false; removeSuperTris=!removeSuperTris; }
#endif

	//
	// Remove triangles with supertriangle vertices
	// These are triangles which have a vertex number greater than nv
	//
	if (1 && removeSuperTris)
	for(int i=0; i<numTris; i++) 
	{
		if(tris[i].i0 >= totalVerts || tris[i].i1 >= totalVerts || tris[i].i2 >= totalVerts) 
		{
			tris[i] = tris[numTris-1]; // Remove this triangle from the list
			numTris--;
			i--;

			//tris[i].valid = false;
		}
	}

	DBG_ASSERT(numTris>0);

	//
	// Check if any of our triangle edges are intersecting our contour
	// edges, and hence we have to modify those triangles
	//
	static bool enableClipping = true;
	extern bool g_keys[0xffff];
	if (g_keys[VK_CONTROL]){ enableClipping=!enableClipping; g_keys[VK_CONTROL]=false; }

	if (1 && enableClipping)
	{
		//
		// First determine which contour lines are clipping triangles
		// 
		int start = 0;
		for (int kk=0; kk<numContours; kk++)
		{
			int numPoints = numPntsInEachContour[kk];
			DBG_ASSERT(numPoints>0);

			int nextI = start;
			for (int nn=0; nn<numPoints; nn++)
			{
				int fromI = nextI ; //start + nn;
				int toI   = start + (nn+1)%numPoints;

				D3DXVECTOR2 from = inVerts[fromI];
				D3DXVECTOR2 to   = inVerts[toI];

				// Check for points which are essentially the same
				// point
				const float deltaFromTo = D3DXVec2LengthSq( &(from - to) );
				static float minDelta = 0.00000001f;
				if (deltaFromTo < minDelta )
				{
					//////DBG_HALT; // can cause problems - to investiage
					continue;
				}
				nextI = toI;

				bool validLine = false;
				for (int i=0; i<numTris; i++)
				{
					int triCorns[3] = { tris[i].i0, tris[i].i1, tris[i].i2 };
					for (int tc=0; tc<3; tc++)
					{
						int cor0 = triCorns[tc];
						int cor1 = triCorns[(tc+1)%3];

						if ( (cor0==fromI && cor1==toI) ||
							 (cor1==fromI && cor0==toI) )
						{
							validLine = true;
							break;
						}
					}

					if (validLine) break;
				}

				if (validLine) continue;

				// If we got here, then this line is intersecting
				// some triangles - so we hvae to determine which triangles
				// and correct them

				// 
				// Work out if any of our triangles cross our contour edges, and mark
				// them as needing removing and correcting
				//
				const int MAX_TRIS_REMOVE = 5000;
				int trisToRemove[MAX_TRIS_REMOVE] = {-1};
				memset(trisToRemove, -1, sizeof(trisToRemove));
				int numTrisToRemove = 0;
				for (int i=0; i<numTris; i++)
				{
					if (!tris[i].valid) continue;

					int corners[3] = { tris[i].i0, tris[i].i1, tris[i].i2 };
					for (int c=0; c<3; c++)
					{
						int cur = corners[c];
						int nex = corners[(c+1)%3];

						if (cur==fromI || cur==toI || nex==fromI || nex==toI) continue;

						D3DXVECTOR2 p0 = verts[ cur ];
						D3DXVECTOR2 p1 = verts[ nex ];

						bool hit =
						LineLineIntersect(from, to, p0, p1, NULL);

						if (hit)
						{
							DBG_ASSERT(numTrisToRemove<(MAX_TRIS_REMOVE-50));
							trisToRemove[numTrisToRemove] = i;
							numTrisToRemove++;
							break;
						}
					}
				}

				if (numTrisToRemove<=1) continue;

				//DBG_ASSERT(numTrisToRemove!=0);

				// Now we know which triangles cross our lines, we have to determine
				// which are on one side and which are on the other side
				// and re-triangulate them

				static bool showRemoveTris = false;
				if (g_keys[VK_INSERT]) { g_keys[VK_INSERT]=false; showRemoveTris=!showRemoveTris; }
#if 0
				// Debug draw which triangles are going to be removed
				if (showRemoveTris)
				for (int i=0; i<numTrisToRemove; i++)
				{
					int i0 = tris[ trisToRemove[i] ].i0;
					int i1 = tris[ trisToRemove[i] ].i1;
					int i2 = tris[ trisToRemove[i] ].i2;

					extern float				g_zoom;
					extern D3DXVECTOR2			g_off;
					float s =					g_zoom;
					extern LPDIRECT3DDEVICE9	g_pD3DDevice;


					D3DXVECTOR2 triPnts[3] = {	verts[i0]*s + g_off, 
												verts[i1]*s + g_off, 
												verts[i2]*s + g_off };

					triPnts[0] = triPnts[0] + (triPnts[1] - triPnts[0])*0.05f;
					triPnts[1] = triPnts[1] + (triPnts[2] - triPnts[1])*0.05f;
					triPnts[2] = triPnts[2] + (triPnts[0] - triPnts[2])*0.05f;

					DebugDrawTri2D(g_pD3DDevice, triPnts, 0xee0000ff);
				}
				//continue;
#endif
				{
					// Extract all the points from our triangle
					const int MAX_TEMP_EDGES = 15000;
					int tempEdges[MAX_TEMP_EDGES] = {-1};
					memset(tempEdges, -1, sizeof(tempEdges));
					int numTempEdges   = 0;

					for (int i=0; i<numTrisToRemove; i++)
					{
						int indx  = trisToRemove[i];

						int i0 = tris[indx].i0;
						int i1 = tris[indx].i1;
						int i2 = tris[indx].i2;

						DBG_ASSERT(numTempEdges+3<MAX_TEMP_EDGES);

						tempEdges[numTempEdges]=i0;		numTempEdges++;
						tempEdges[numTempEdges]=i1;		numTempEdges++;
						tempEdges[numTempEdges]=i2;		numTempEdges++;
					}

					D3DXVECTOR2 norm = D3DXVECTOR2( from.y - to.y,
												    to.x   - from.x);
					D3DXVec2Normalize(&norm, &norm);

					static float eachSideOffset = 0.005f;

					D3DXVECTOR2 pStartTop  = from + norm*eachSideOffset;
					D3DXVECTOR2 pBottomTop = from - norm*eachSideOffset;

					float dTop    = (pStartTop.x  * norm.x) + (pStartTop.y * norm.y);
					float dBottom = (pBottomTop.x * norm.x) + (pBottomTop.y * norm.y);

					const int MAX_TEMP_INDS			= MAX_TEMP_EDGES;
					int top[MAX_TEMP_INDS]			= {-1};
					int bottom[MAX_TEMP_INDS]		= {-1};
					int numTop						= 0;
					int numBottom					= 0;



					// Remove any duplicates
					for (int i=0; i<numTempEdges; i++)
					{
						for (int f=i+1; f<numTempEdges; f++)
						{
							if (tempEdges[i] == tempEdges[f])
							{
								tempEdges[f] = tempEdges[numTempEdges-1];
								numTempEdges--;
								i--;
							}
						}
					}

					// Work out which points are on which each side of our intersecting
					// line
					for (int i=0; i<numTempEdges; i++)
					{
						int indx = tempEdges[i];

						D3DXVECTOR2 pnt = verts[indx];

						float dT = (pnt.x * norm.x) + (pnt.y * norm.y);
						if (dT < dTop) // Top
						{
							DBG_ASSERT(numTop<MAX_TEMP_INDS-1);
							top[numTop] = indx;
							numTop++;
						}

						float dB = (pnt.x * norm.x) + (pnt.y * norm.y);
						if (dB > dBottom) // Bottom
						{
							DBG_ASSERT(numBottom<MAX_TEMP_INDS-1);
							bottom[numBottom] = indx;
							numBottom++;
						}
					}
#if 0
					if (showRemoveTris)				
					for (int i=0; i<numTop; i++)
					{
						extern float		g_zoom;
						extern D3DXVECTOR2	g_off;
						float s	=			g_zoom;
						extern LPDIRECT3DDEVICE9	g_pD3DDevice;

						D3DXVECTOR2 pnt = verts[ top[i] ]*s + g_off + D3DXVECTOR2(3.0f, 0);
						DebugDrawCircle2D(g_pD3DDevice, pnt, 8.0f, 0xffff0000);

					}
					if (showRemoveTris)
					for (int i=0; i<numBottom; i++)
					{
						extern float		g_zoom;
						extern D3DXVECTOR2	g_off;
						float s	=			g_zoom;
						extern LPDIRECT3DDEVICE9	g_pD3DDevice;

						D3DXVECTOR2 pnt = verts[ bottom[i] ]*s + g_off;
						DebugDrawCircle2D(g_pD3DDevice, pnt, 8.0f, 0xff00ff00);

					}
#endif

					int* sideTris[2]   = {top, bottom};
					int  numOnSide[2]  = {numTop, numBottom};

					for (int i=0; i<2; i++)
					{
						int* sideT     = sideTris[i];
						int  numSideT  = numOnSide[i];

						DBG_ASSERT(numTris+1<maxTris);
						
						D3DXVECTOR2 tmpV[500];
						memset(tmpV, -1, sizeof(tmpV));
						for (int kn=0; kn<numSideT; kn++)
						{
							DBG_ASSERT(kn<499);
							tmpV[kn] = verts[ sideT[kn] ];
						}

						int  numFound = 0;
						
						const int MAX_SUB_TRIS = 5000;
						int* tmpOutTriInds = new int[MAX_SUB_TRIS];
						memset(tmpOutTriInds, -1, MAX_SUB_TRIS*sizeof(int));
						TriangulatePoints(tmpV, numSideT, MAX_SUB_TRIS, &numFound, tmpOutTriInds);
						
						for (int kn=0; kn<numFound; kn++)
						{
							tris[numTris].i0 = sideT[ tmpOutTriInds[kn*3 + 0] ];
							tris[numTris].i1 = sideT[ tmpOutTriInds[kn*3 + 1] ];
							tris[numTris].i2 = sideT[ tmpOutTriInds[kn*3 + 2] ];

							//To-Do - Remove invalid tris
							numTris++;
						}

						delete[] tmpOutTriInds;
					}

#if defined(DEBUG_STEP) && 1
					{
						extern float		g_zoom;
						extern D3DXVECTOR2	g_off;
						float s = g_zoom;
						extern LPDIRECT3DDEVICE9	g_pD3DDevice;

						D3DXVECTOR2 dbgOff = D3DXVECTOR2(0.0f, 1.0f);
						DebugDrawArrowLine2D(g_pD3DDevice, (from+dbgOff)*s+g_off, (to+dbgOff)*s+g_off, 0xff00ffff);					
					}
#endif
					
					for (int ii=0; ii<numTrisToRemove; ii++)
					{
						int tr = trisToRemove[ii];

						tris[tr] = tris[numTris-1];
						numTris--;
					}

				}


			}

			start+=numPoints;
		}


	
	}// If Clipping Enabled

	
	// Check for duplicate triangles
	/*
	for (int i=0; i<numTris; i++)
	{
		for (int k=i+1; k<numTris; k++)
		{
			if (tris[i].i0==tris[k].i0 &&
				tris[i].i1==tris[k].i1 &&
				tris[i].i2==tris[k].i2)
			{
				tris[i] = tris[numTris-1];
				numTris--;
				i--;
				break;
			}
		}
	}
	*/


	//
	// Determine if our triangle is visible, either its a hole or not
	//
	static bool doHoles = true;
	if (g_keys[VK_SHIFT]){ doHoles=!doHoles; g_keys[VK_SHIFT]=false; }

	if (doHoles)
	for (int i=0; i<numTris; i++)
	{
		D3DXVECTOR2 p0 = verts[ tris[i].i0 ];
		D3DXVECTOR2 p1 = verts[ tris[i].i1 ];
		D3DXVECTOR2 p2 = verts[ tris[i].i2 ];

		D3DXVECTOR2 centre = (p0 + p1 + p2)/3.0f;
		int numHits = 0;

		int start = 0;
		for (int k=0; k<numContours; k++)
		{
			int numPoints = numPntsInEachContour[k];
			DBG_ASSERT(numPoints>0);

			for (int n=0; n<numPoints; n++)
			{
				D3DXVECTOR2 from = inVerts[start + n];
				D3DXVECTOR2 to   = inVerts[start + (n+1)%numPoints];

				if (centre.y < to.y)
				{
					if (from.y <= centre.y)
					{
						if ((centre.y-from.y)*(to.x-from.x) > (centre.x-from.x)*(to.y-from.y))
						{
							numHits++;
						}
					}
				}
				else if (centre.y < from.y)
				{
					if ((centre.y-from.y)*(to.x-from.x) < (centre.x-from.x)*(to.y-from.y))
					{
						numHits++;
					}
				}

			}
			start += numPoints;
		}

		if (numHits%2 == 0)
		{
			tris[i] = tris[numTris-1];
			numTris--;
			i--;
		}
	}
/*
if (1 && doHoles)
	for (int i=0; i<numTris; i++)
	{
		if (!tris[i].valid) continue;

		D3DXVECTOR2 p0 = verts[ tris[i].i0 ];
		D3DXVECTOR2 p1 = verts[ tris[i].i1 ];
		D3DXVECTOR2 p2 = verts[ tris[i].i2 ];

		D3DXVECTOR2 centre = (p0 + p1 + p2)/3.0f;
		int numHits = 0;

		int start = 0;
		for (int k=0; k<numContours; k++)
		{
			int numPoints = numPntsInEachContour[k];
			DBG_ASSERT(numPoints>0);

			for (int n=0; n<numPoints; n++)
			{
				D3DXVECTOR2 from = inVerts[start + n];
				D3DXVECTOR2 to   = inVerts[start + (n+1)%numPoints];
				
#ifdef DEBUG_STEP
				static bool drawRay = false;
				if (drawRay)
				{
					extern float		g_zoom;
					extern D3DXVECTOR2	g_off;
					float s = g_zoom;
					D3DXVECTOR2 pnts[] = { centre*s + g_off, (centre+D3DXVECTOR2(9000.0f,-9000.0f))*s + g_off };
					extern LPDIRECT3DDEVICE9	g_pD3DDevice;
					DebugDrawLineList2D(g_pD3DDevice, pnts, 2, 0xfff00fff);
				}
#endif

				bool hit =
				LineLineIntersect(from, to, centre, centre+D3DXVECTOR2(9000.0f,-9000.0f), NULL, -0.0000000001f);

				if (hit)
				{
					numHits++;
				}
			}
			start += numPoints;
		}

		if (numHits%2 == 0)
		{
			tris[i] = tris[numTris-1];
			numTris--;
			i--;

			//tris[i].valid=false;
		}
	}
	*/


#if defined (DEBUG_STEP) && 1
	if (g_doDebugStep)
	{
		extern float		g_zoom;
		extern D3DXVECTOR2	g_off;
		float s = g_zoom;
		extern LPDIRECT3DDEVICE9	g_pD3DDevice;

		POINT pnt;
		GetCursorPos(&pnt);
		extern HWND	g_hWnd;
		::ScreenToClient(g_hWnd, &pnt);
		D3DXVECTOR2 curPos = D3DXVECTOR2((float)pnt.x, (float)pnt.y);
		DebugDrawCircle2D(g_pD3DDevice, curPos, 3.0f, 0xCC430Fff);




		D3DXVECTOR2 testV = verts[ (g_debugStep+1)%totalVerts ];

		DebugDrawCircle2D(g_pD3DDevice, testV*s + g_off, 2.0f, 0xBB3300ff);

		for (int i=0; i<numTris; i++)
		{
			D3DXVECTOR2 p0 = verts[ tris[i].i0 ];
			D3DXVECTOR2 p1 = verts[ tris[i].i1 ];
			D3DXVECTOR2 p2 = verts[ tris[i].i2 ];

			D3DXVECTOR2 triPntsX[3] = {	p0*s + g_off, 
										p1*s + g_off, 
										p2*s + g_off};

			bool inTri = IsPointInTriangle(triPntsX[0], triPntsX[1], triPntsX[2], curPos);
			if (inTri)
			{
				float xc, yc, r;
				CircumCircle(  curPos.x, curPos.y,	// Test Vert
							   triPntsX[0].x,    triPntsX[0].y,		// 3 Corners of the triangle
							   triPntsX[1].x,    triPntsX[1].y, 
							   triPntsX[2].x,    triPntsX[2].y, 
							   xc, yc, r);
				DebugDrawCircle2D(g_pD3DDevice, D3DXVECTOR2(xc,yc), r, 0xffff0000);

			}

			if(tris[i].i0 >= totalVerts || tris[i].i1 >= totalVerts || tris[i].i2 >= totalVerts) 
			{
				continue;
			}

			float xc, yc, r;
			bool inside = CircumCircle(testV.x, testV.y,// Test Vert
									   p0.x,    p0.y,	// 3 Corners of the triangle
								       p1.x,    p1.y, 
								       p2.x,    p2.y, 
								       xc, yc, r);		// Results

			if (inside)
			{
				extern float		g_zoom;
				extern D3DXVECTOR2	g_off;
				float s = g_zoom;
				extern LPDIRECT3DDEVICE9	g_pD3DDevice;

				D3DXVECTOR2 pntt = D3DXVECTOR2(xc, yc);
				pntt = pntt*s + g_off;
				r = r*s;
				unsigned int col = 0xEFE1FF44;
				DebugDrawCircle2D(g_pD3DDevice, pntt, r, 0xEE448844);

				static bool enableNextTri = false;
				if (g_keys[VK_PAUSE])
				{
					g_keys[VK_PAUSE] = false;
					enableNextTri = !enableNextTri;
				}

				if (enableNextTri)
				{
					D3DXVECTOR2 triPnts[3] = {	p0*s + g_off, 
												p1*s + g_off, 
												p2*s + g_off};

					DebugDrawCircle2D(g_pD3DDevice, triPnts[0], 8, 0xff00ff00);
					DebugDrawCircle2D(g_pD3DDevice, triPnts[1], 8, 0xff00ff00);
					DebugDrawCircle2D(g_pD3DDevice, triPnts[2], 8, 0xff00ff00);

					DebugDrawTri2D(g_pD3DDevice, triPnts, 0xEE0ff000);
				}

			}
		}
	}

#endif 



#if defined (DEBUG_STEP) && 1 // Debug Draw Our List of Triangles
	for (int i=0; i<numTris; i++)
	{
		if (!tris[i].valid) continue;

		int i0 = tris[i].i0;
		int i1 = tris[i].i1;
		int i2 = tris[i].i2;

		D3DXVECTOR2 v0 = D3DXVECTOR2(verts[i0].x, verts[i0].y);
		D3DXVECTOR2 v1 = D3DXVECTOR2(verts[i1].x, verts[i1].y);
		D3DXVECTOR2 v2 = D3DXVECTOR2(verts[i2].x, verts[i2].y);

		D3DXVECTOR3 pnts[3] = {v0, v1, v2};
		for (int k=0; k<3; k++)
		{
			D3DXVECTOR2 p0 = pnts[k];
			D3DXVECTOR2 p1 = pnts[(k+1)%3];

			extern float		g_zoom;
			extern D3DXVECTOR2	g_off;
			float s = g_zoom;

			D3DXVECTOR2 pnts[] = { p0*s + g_off, p1*s + g_off };

			unsigned char r = (i * 10) % 255;
			unsigned char g = (i * 90) % 255;
			unsigned char b = (i * 50) % 255;
			unsigned int col = (0xff<<24) | (r<<16) | (g<8) | (b);
			extern LPDIRECT3DDEVICE9	g_pD3DDevice;

			static bool drawLines = true;
			extern bool	g_keys[0xffff];
			if (g_keys[VK_TAB]) { g_keys[VK_TAB]=false; drawLines=!drawLines; }
			if (drawLines)
			{
				DebugDrawLineList2D(g_pD3DDevice, pnts, 2, col);
			}
		}

		{
			extern float		g_zoom;
			extern D3DXVECTOR2	g_off;
			float s = g_zoom;
			extern LPDIRECT3DDEVICE9	g_pD3DDevice;

			D3DXVECTOR2 triPnts[3] = {	v0*s + g_off, 
										v1*s + g_off, 
										v2*s + g_off };

			DebugDrawTri2D(g_pD3DDevice, triPnts, 0x33ff0000);
		}
	}
#endif

#if defined (DEBUG_STEP)
	{
	// Debug draw contours that where passed in with arrows
	int start = 0;
	for (int i=0; i<numContours; i++)
	{
		int numPoints = numPntsInEachContour[i];
		for (int k=0; k<numPoints; k++)
		{
			extern float		g_zoom;
			extern D3DXVECTOR2	g_off;
			float s = g_zoom;
			extern LPDIRECT3DDEVICE9	g_pD3DDevice;

			D3DXVECTOR2 from = inVerts[start + k]*s + g_off;
			D3DXVECTOR2 to   = inVerts[start + (k+1)%numPoints]*s + g_off;
			DebugDrawArrowLine2D(g_pD3DDevice, from, to, 0xAAffee00);
		}
		start += numPoints;
	}
	}
#endif


	int count = 0;
	(*outNumTris) = 0;
	for (int i=0; i<numTris; i++)
	{
		if (!tris[i].valid) continue;

		DBG_ASSERT(count+3 < maxOutTris);
		if (count+3 >= maxOutTris) break;

		int i0 = tris[i].i0;
		int i1 = tris[i].i1;
		int i2 = tris[i].i2;

		if (i0>=totalVerts || i1>=totalVerts || i2>=totalVerts) continue;

	#ifndef DEBUG_STEP
		DBG_ASSERT(i0>=0 && i0<totalVerts);
		DBG_ASSERT(i1>=0 && i1<totalVerts);
		DBG_ASSERT(i2>=0 && i2<totalVerts);
	#endif

		outTriIndices[count]=i0;		count++;
		outTriIndices[count]=i1;		count++;
		outTriIndices[count]=i2;		count++;
		(*outNumTris) += 1;
	}

	delete[] verts;
	delete[] tris;
	delete[] edges;


	//(*outNumTris) = 0;

	return 0;
}



/******************************************************************************/



